import json
import re
import os
import pandas as pd
import numpy as np
from collections import OrderedDict,namedtuple
import sys,traceback

class SensorQC:
    def __init__(self,base_path="",**kwargs):
        self.base_path=base_path
    
    def readcsv_pd(self,csvname):
        #temp_pd=pd.read_csv(os.path.join(self.base_path, csvname))
        temp_pd=pd.read_csv(self.base_path + csvname)
        return temp_pd

class SensorQCVac(SensorQC):
    def __init__(self,vacuum_offset=14.7,off_state_threshold=6.7,on_state_threshold=5,**kwargs):
        super().__init__(**kwargs)
        self.vacuum_offset=vacuum_offset
        self.off_state_threshold=off_state_threshold
        self.on_state_threshold=on_state_threshold
        self.input_json=kwargs["input_json"]
        self.vac_sensor_indicator=kwargs["vac_sensor_indicator"]
        self.ps_sensor_indicator=kwargs["ps_sensor_indicator"]

    def transform_data(self):
        csv_available_indexes=[index for index,value in enumerate(self.input_json) if value.csvAvailable==True]   #this returns the indexes as list where csv is available
        expected_vacuum=[self.input_json[i].data.content for i in csv_available_indexes]                          #this returns the data content from json as a list of string
        expected_vacuum_reversed = expected_vacuum[::-1]
        self.csv_expected_vacuum_dict = dict(zip(csv_available_indexes,expected_vacuum))
        self.expected_vacuum_unique=list(OrderedDict.fromkeys(expected_vacuum_reversed))  
        self.expected_vacuum_unique.reverse()                           
        #print(self.expected_vacuum_unique)                                         #unique data content returned as a list of string
        self.number_of_unique_vac_values=len(self.expected_vacuum_unique)

    def get_calibration_attempts(self):
        #here we return the latest 2 calibration attempts
        last_two_unique_vacuum_values=sorted(self.expected_vacuum_unique[self.number_of_unique_vac_values-2:self.number_of_unique_vac_values]) 
        off_state_index= [key for key,value in self.csv_expected_vacuum_dict.items() if value == max(last_two_unique_vacuum_values)]
        on_state_index= [key for key,value in self.csv_expected_vacuum_dict.items() if value == min(last_two_unique_vacuum_values)]
        #======the below 2 lines are storing the off_state and on_state values obtained======
        off_state_calibration_attempt = self.input_json[off_state_index[-1]]
        on_state_calibration_attempt = self.input_json[on_state_index[-1]]
        self.off_state_data_0=self.readcsv_pd(off_state_calibration_attempt.csvPath)
        self.on_state_data_0=self.readcsv_pd(on_state_calibration_attempt.csvPath)
        self.off_state_vacuum=float(last_two_unique_vacuum_values[1])             #result of off_state vacuum stored
        self.on_state_vacuum=float(last_two_unique_vacuum_values[0])              #result of on_state vacuum stored

    def get_state_data(self):
        if self.ps_sensor_indicator:
            #print(on_state_vacuum)
            self.off_state_vacuum = self.off_state_vacuum - self.vacuum_offset
            self.on_state_vacuum = self.on_state_vacuum - self.vacuum_offset
            self.off_state_data=self.off_state_data_0["ps"].where(self.off_state_data_0["ps"]<=50) 
            self.on_state_data=self.on_state_data_0["ps"].where(self.on_state_data_0["ps"]<=50) 
        #========for vacuum======
        if self.vac_sensor_indicator:
            self.off_state_data=self.off_state_data_0["vac"].where(self.off_state_data_0["vac"]<=20) #.apply(lambda x: x if x<=20 else 0)
            self.on_state_data=self.on_state_data_0["vac"].where(self.on_state_data_0["vac"]<=20) #.apply(lambda x: x if x<=20 else 0)            

    def get_vaccuum_values(self):
        self.off_state_observed_vacuum=round(self.off_state_data.median(),2)
        self.on_state_observed_vacuum=round(self.on_state_data.median(),2)
        if self.ps_sensor_indicator:
            self.off_state_observed_vacuum=self.off_state_observed_vacuum-self.vacuum_offset
            self.on_state_observed_vacuum=self.on_state_observed_vacuum-self.vacuum_offset 
        self.vacuum_values={"off_state_vacuum" : self.off_state_vacuum, "off_state_observed_vacuum" : self.off_state_observed_vacuum ,
                    "on_state_vacuum" : self.on_state_vacuum, "on_state_observed_vacuum" : self.on_state_observed_vacuum}    
    
    def qc_compute(self):
        #=======qc indicators============= #======0 means pass and 1 means fail and 2.2 psi is cutoff
        qc_indicators={}
        qc_indicators["off"]=0 if abs(self.off_state_observed_vacuum - self.off_state_vacuum ) <= self.off_state_threshold else 1
        qc_indicators["on"]=0 if abs(self.on_state_observed_vacuum - self.on_state_vacuum ) <= self.on_state_threshold else 1
        self.combined_indicator=0 if qc_indicators["off"]+qc_indicators["on"]==0 else 1
        #......... Some information regarding sensor qc ..........    #================Creating reason information======================#....... Off state sensor qc message ........
        for indicator in qc_indicators:
            status = "failed" if qc_indicators[indicator]==1 else "passed"
            verdict="more" if qc_indicators[indicator]==1 else "less"
            state_vacuum=self.vacuum_values[indicator+"_state_vacuum"];state_observed_vacuum=self.vacuum_values[indicator+"_state_observed_vacuum"];thres_string=f"self.{indicator}_state_threshold";state_threshold=eval(thres_string)
            qc_indicators[indicator]=f"Sensor QC for \"{indicator.upper()} state {status}\" as difference between {indicator.upper()} State Vaccuum ({state_vacuum}) and Observed {indicator.upper()} State Vacuum ({state_observed_vacuum}) is {verdict} than {indicator.upper()} State Threshold : {state_threshold}"
        self.sensor_pass_fail_reason=qc_indicators["off"]+" and "+qc_indicators["on"]
        
    def process_qc(self):
        self.transform_data()
        if self.number_of_unique_vac_values < 2:
            return{"insufficient_data_indicator": 1}
        self.get_calibration_attempts()
        self.get_state_data()
        self.get_vaccuum_values()
        self.qc_compute()
        if self.combined_indicator!=0:
            return { "insufficient_data_indicator":0, "result":"Failed", "sensor_pass_fail_reason_list":self.sensor_pass_fail_reason,"vacuum_values":self.vacuum_values}
        return {"insufficient_data_indicator":0, "result":"Passed", "vacuum_values":self.vacuum_values, "sensor_pass_fail_reason_list": self.sensor_pass_fail_reason}
        
            
class SensorQCPF(SensorQC):
    def __init__(self,off_state_voltage_threshold=10,off_state_current_threshold=2,on_state_voltage_threshold=30,on_state_current_threshold=100,**kwargs):
        super().__init__(**kwargs)
        self.input_json=kwargs["input_json"]
        self.qc_information_api_data_pf=kwargs["qc_information_api_data_pf"]
        self.parameters_pf_called=kwargs["parameters_pf_called"]
        self.off_state_voltage_threshold=off_state_voltage_threshold
        self.off_state_current_threshold=off_state_current_threshold
        self.on_state_voltage_threshold=on_state_voltage_threshold
        self.on_state_current_threshold=on_state_current_threshold
    
    def transform_data(self):
        csv_available_indexes=[index for index,value in enumerate(self.input_json) if value.csvAvailable==True]   #this returns the indexes as list where csv is available
        expected_current=[i.split(",") for i in [self.input_json[i].data.content for i in csv_available_indexes]]
        expected_current_float=[list(map(float, sub_list)) for sub_list in expected_current]                   #this converts the string integers to float 
        for i in expected_current_float:
            if len(i)!=6:
                expected_current_float.remove(i)
        keys=["v1","i1","v2","i2","v3","i3"]
        expected_current_voltage_values={x:list(y) for x,y in zip(keys, zip(*expected_current_float))}                   #dicyionary created and later moving this to dataframe
        self.expected_current_voltage_values_df=pd.DataFrame(expected_current_voltage_values)
        self.expected_current_voltage_values_df.index = csv_available_indexes
        self.expected_current_voltage_values_df_unique_0=self.expected_current_voltage_values_df.drop_duplicates(keep="last")     #removing duplicates
        self.expected_current_voltage_values_df_unique_0=self.expected_current_voltage_values_df_unique_0.dropna()                #removing rows with NaN
        self.expected_current_voltage_values_df_unique=self.expected_current_voltage_values_df_unique_0.tail(2)

    def get_state_index(self):
        selected_indices=self.expected_current_voltage_values_df_unique.index.values.astype(int)                           #index in list datatype
        min_i1=self.expected_current_voltage_values_df["i1"].min()
        self.off_state_index=selected_indices[self.expected_current_voltage_values_df_unique["i1"]==min_i1]
        self.on_state_index=selected_indices[self.expected_current_voltage_values_df_unique["i1"]!=min_i1]
        
    def get_state_data(self):
        self.off_state_voltage=self.expected_current_voltage_values_df_unique[["v1","v2","v3"]].loc[self.off_state_index]
        self.off_state_current=self.expected_current_voltage_values_df_unique[["i1","i2","i3"]].loc[self.off_state_index]
        self.on_state_voltage=self.expected_current_voltage_values_df_unique[["v1","v2","v3"]].loc[self.on_state_index]
        self.on_state_current=self.expected_current_voltage_values_df_unique[["i1","i2","i3"]].loc[self.on_state_index]
        observed_off_state_list = self.Formatting_pf_raw_data(self.off_state_index)
        self.observed_off_state_voltage = np.array(observed_off_state_list["voltage"])
        self.observed_off_state_current = np.array(observed_off_state_list["current"])
        observed_on_state_list = self.Formatting_pf_raw_data(self.on_state_index)
        self.observed_on_state_voltage = np.array(observed_on_state_list["voltage"])
        self.observed_on_state_current = np.array(observed_on_state_list["current"])

    def get_calibration_attempts(self):
        #===========modifying raw values===========
        try:
            self.previous_qc_calibration_parameters=self.qc_information_api_data_pf["calibrationParam"]["calibration_parameters"]
        except:
            self.previous_qc_calibration_parameters=[]

    def calculate_phase_voltages(self):
        try:
            self.voltage_a = np.array([float(self.previous_qc_calibration_parameters["voltage_phase1_a"]["v1"]), float(self.previous_qc_calibration_parameters["voltage_phase2_a"]["v2"]), float(self.previous_qc_calibration_parameters["voltage_phase3_a"]["v3"])])
            self.voltage_b = np.array([float(self.previous_qc_calibration_parameters["voltage_phase1_b"]["v1"]), float(self.previous_qc_calibration_parameters["voltage_phase2_b"]["v2"]), float(self.previous_qc_calibration_parameters["voltage_phase3_b"]["v3"])])
            self.current_a = np.array([float(self.previous_qc_calibration_parameters["current_phase1_a"]["i1"]), float(self.previous_qc_calibration_parameters["current_phase2_a"]["i2"]), float(self.previous_qc_calibration_parameters["current_phase3_a"]["i3"])])
            self.current_b = np.array([float(self.previous_qc_calibration_parameters["current_phase1_b"]["i1"]), float(self.previous_qc_calibration_parameters["current_phase2_b"]["i2"]), float(self.previous_qc_calibration_parameters["current_phase3_b"]["i3"])])
        except:
            self.voltage_a = np.resize( 1, 3 ) ; self.voltage_b = np.resize( 0, 3 ) ; self.current_a = np.resize( 1, 3 ) ; self.current_b = np.resize( 0, 3 )

    def calculate_observed_state_voltages(self):
        self.modified_observed_off_state_voltage=self.observed_off_state_voltage/self.voltage_a
        self.modified_observed_on_state_voltage=(self.observed_on_state_voltage-self.voltage_b)/self.voltage_a
        self.modified_observed_off_state_current=self.observed_off_state_current/self.current_a
        self.modified_observed_on_state_current=(self.observed_on_state_current-self.current_b )/self.current_a

    def correct_thresholds(self):
        if self.parameters_pf_called["isostat"]=="Yes":
            self.off_state_voltage_threshold = 2
        self.batch_cttpe=self.parameters_pf_called["cttype"]
        if self.batch_cttpe is None or self.batch_cttpe in (100,300):
            self.off_state_current_threshold=1
        if self.parameters_pf_called["cttype"] is None or self.parameters_pf_called["cttype"]==100:
            self.on_state_current_threshold=30

    def set_state_indicator(self):
        self.qc_status={"Voltage":{"OFF":0,"ON":0},
                         "Current":{"OFF":0,"ON":0}}
        self.qc_status["Voltage"]["OFF"]= 0 if (abs(self.off_state_voltage-self.modified_observed_off_state_voltage)<=self.off_state_voltage_threshold).all(axis=None) else 1
        self.qc_status["Current"]["OFF"]= 0 if (abs(self.off_state_current-self.modified_observed_off_state_current)<=self.off_state_current_threshold).all(axis=None) else 1
        self.qc_status["Voltage"]["ON"] = 0 if (abs(self.on_state_voltage - self.modified_observed_on_state_voltage)<=self.on_state_voltage_threshold).all(axis=None) else 1
        self.qc_status["Current"]["ON"] = 0 if (abs(self.on_state_current - self.modified_observed_on_state_current)<=self.on_state_current_threshold).all(axis=None) else 1
        self.combined_indicator = 0 if self.qc_status["Voltage"]["OFF"] + self.qc_status["Current"]["OFF"] + self.qc_status["Voltage"]["ON"]+ self.qc_status["Current"]["ON"]  == 0 else 1
        current_pass_indicator = "" if self.qc_status["Current"]["OFF"] + self.qc_status["Current"]["ON"]  == 0 else " Current Sensor Failed "
        voltage_pass_indicator = "" if self.qc_status["Voltage"]["OFF"] + self.qc_status["Voltage"]["ON"]== 0 else " Voltage Sensor Failed "
        self.failure_reason = current_pass_indicator+" "+voltage_pass_indicator
        
    def form_qc_status(self, val):
        return "PASSED" if val == 0 else "FAILED"

    def set_qc_criteria(self):
        #......... Some information regarding sensor qc ..........
        self.verdict={}
        self.verdict["PASSED"]="LESS";self.verdict["FAILED"]="GREATER"
        self.qc_status={"Voltage":{"OFF":self.form_qc_status(self.qc_status["Voltage"]["OFF"]),"ON":self.form_qc_status(self.qc_status["Voltage"]["ON"])},
                         "Current":{"OFF":self.form_qc_status(self.qc_status["Current"]["OFF"]),"ON":self.form_qc_status(self.qc_status["Current"]["ON"])}}
    
    def evaluate_observed_values(self):
        #string values to return
        self.off_state_voltage["combined_voltage"]=self.off_state_voltage["v1"].astype(str)+" , "+self.off_state_voltage["v2"].astype(str)+" ,"+self.off_state_voltage["v3"].astype(str)
        self.on_state_voltage["combined_voltage"]=self.on_state_voltage["v1"].astype(str)+" , "+self.on_state_voltage["v2"].astype(str)+" ,"+self.on_state_voltage["v3"].astype(str)
        self.modified_observed_off_state_voltage_combined=",".join(map(str, self.modified_observed_off_state_voltage))
        self.modified_observed_on_state_voltage_combined= ",".join(map(str, self.modified_observed_on_state_voltage))
        self.off_state_current["combined_current"]=self.off_state_current["i1"].astype(str)+" , "+self.off_state_current["i2"].astype(str)+" ,"+self.off_state_current["i3"].astype(str)
        self.on_state_current["combined_current"]=self.on_state_current["i1"].astype(str)+" , "+self.on_state_current["i2"].astype(str)+" ,"+self.on_state_current["i3"].astype(str)
        self.modified_observed_off_state_current_combined= ",".join(map(str, self.modified_observed_off_state_current))                
        self.modified_observed_on_state_current_combined= ",".join(map(str, self.modified_observed_on_state_current))
   
    def qc_compute(self):
        for variable in self.qc_status:
            for state in self.qc_status[variable]:
                lstate=state.lower();lvariable=variable.lower()
                form_string=f"self.{lstate}_state_{lvariable}['combined_{lvariable}'].values[0]";combined_variable=eval(form_string)
                form_string=f"self.modified_observed_{lstate}_state_{lvariable}_combined";modified_observed_state=eval(form_string)
                form_string=f"self.{lstate}_state_{lvariable}_threshold";state_threshold=eval(form_string)
                self.qc_status[variable][state]=f"Sensor QC for \"{variable} {state} state {self.qc_status[variable][state]}\" as difference between {state} State {variable} ( {combined_variable}) and Observed {state} State {variable} ( { str(modified_observed_state) }) is {self.verdict[self.qc_status[variable][state]]} than {state} State {variable} Threshold : {str(state_threshold)}"
        self.sensor_pass_fail_reason=self.qc_status["Voltage"]["OFF"]+ " ........... "+ self.qc_status["Current"]["OFF"]+ " ........... "+ self.qc_status["Voltage"]["ON"] +" ................ "+self.qc_status["Current"]["ON"]
 
    def form_qc_result(self):
        qc_result={"insufficient_data_indicator":0, "result": "Passed", "voltage_values": {"modified_observed_on_state_voltage": self.modified_observed_on_state_voltage,
                                                                                    "modified_observed_off_state_voltage": self.modified_observed_off_state_voltage,
                                                                                    "off_state_voltage": np.array(list(eval(np.array(self.off_state_voltage["combined_voltage"])[0]))),
                                                                                    "on_state_voltage": np.array(list(eval(np.array(self.on_state_voltage["combined_voltage"])[0])))},
            "current_values": {"modified_observed_on_state_current": self.modified_observed_on_state_current, "modified_observed_off_state_current": self.modified_observed_off_state_current,
                                "off_state_current": np.array(list(eval(np.array(self.off_state_current["combined_current"])[0]))),"on_state_current": np.array(list(eval(np.array(self.on_state_current["combined_current"])[0])))},"sensor_pass_fail_reason_list": self.sensor_pass_fail_reason,
                "previous_calibration_parameters": {"previous_current_a": self.current_a, "previous_current_b":self.current_b,
                                                "previous_voltage_a":self.voltage_a, "previous_voltage_b":self.voltage_b}}
        return qc_result

    def Formatting_pf_raw_data(self,index):            
        pf_raw_data_state_0=self.readcsv_pd(self.input_json[index[0]].csvPath)
        pf_raw_data_state_pre_formatted=pf_raw_data_state_0["pf"][pf_raw_data_state_0["datatype"]==4]        #from the read csv as dataframe only those pf values are being appended as series datatype where datatype column has value equal to 4
        pf_output=pf_raw_data_state_pre_formatted.str.split(",", n=6, expand=True).drop([6],axis=1)          #new df being created with the series and the 6th column dropped
        pf_output.rename(columns={0:"v1", 1:"i1", 2:"v2", 3:"i2", 4:"v3", 5:"i3"}, inplace=True)             #columns renamed
        column_median=pf_output.median()
        return {"voltage":(column_median.v1,column_median.v2,column_median.v3) , "current":(column_median.i1,column_median.i2,column_median.i3)}
        
    def process_qc(self):
        self.transform_data()
        if self.expected_current_voltage_values_df_unique.shape[0]!=2:
            return {"insufficient_data_indicator":1}
        self.get_state_index()
        self.get_state_data()
        self.get_calibration_attempts()
        self.calculate_phase_voltages()
        self.calculate_observed_state_voltages()
        self.correct_thresholds()
        self.set_state_indicator()
        self.set_qc_criteria()
        self.evaluate_observed_values()
        self.qc_compute()
        if self.combined_indicator!=0:
            return {"insufficient_data_indicator":0, "result": "Failed", "sensor_pass_fail_reason_list": self.sensor_pass_fail_reason, "failure_reason":self.failure_reason,"previous_calibration_parameters": {"previous_current_a": self.current_a, "previous_current_b":self.current_b,"previous_voltage_a":self.voltage_a, "previous_voltage_b":self.voltage_b},"voltage_values": {"modified_observed_on_state_voltage": self.modified_observed_on_state_voltage,
                                                                                    "modified_observed_off_state_voltage": self.modified_observed_off_state_voltage,
                                                                                    "off_state_voltage": np.array(list(eval(np.array(self.off_state_voltage["combined_voltage"])[0]))),
                                                                                    "on_state_voltage": np.array(list(eval(np.array(self.on_state_voltage["combined_voltage"])[0])))},"current_values": {"modified_observed_on_state_current": self.modified_observed_on_state_current, "modified_observed_off_state_current": self.modified_observed_off_state_current,
                                "off_state_current": np.array(list(eval(np.array(self.off_state_current["combined_current"])[0]))),"on_state_current": np.array(list(eval(np.array(self.on_state_current["combined_current"])[0])))}}
        return self.form_qc_result()