# Execute (PF): python generate_qc_format.py -p 7778 -m b827ebb7c0c7 -t PF -c log/calibration/csv/ -i 1 -r log/calibration/profile/
# VAC Format: <timestamp>,<sensor_mac>,<pacakge>,<package_index>,<data_type>,<data_type_index>,<data_value_list>
# Execute (VAC): python generate_qc_format.py -p 7778 -m 001ec03f6876 -t VAC -c log/calibration/csv/ -i 1 -r log/calibration/profile/
import sys,os
import zmq
import csv
import datetime
import getopt
import json
import time
from zmq_convert import get_zmq_data_format2
from pynovatec.events import EventPost, EventRegisterListener,EventUnregisterListener,EventQueueShutdown
from qc_events import QCStateChange,SensorDataPoint,SensorDataCollectionComplete,SensorDataBatchCollectionComplete,SensorDataBatchCollectionStarted,SensorDataCollectionError
from qc_events import QCListener

csvpath="log/calibration/csv/"
profilepath="log/calibration/profile/"

def set_csv_path(p):
    global csvpath
    csvpath=p

def set_profile_path(p):
    global profilepath
    profilepath=p


def set_header_list():
    return ["datetime","type","index","datatype","ir","bp","hy","vac","pf","ps","value","x","y","z","var_x","var_y","var_z","mean_x","mean_y","mean_z"]

def get_epoch(time_str):
    return int(datetime.datetime.strptime(time_str,'%Y-%m-%dT%H:%M:%S.%fZ').timestamp())

def get_datetime(epoch):
    return datetime.datetime.fromtimestamp(epoch/1000).strftime('%Y-%m-%dT%H:%M:%SZ')

class SensorCollector:
    start_time=0
    device_state="off"
    collect_on=False
    content="10"
    collect_points=120
    expected_state="off"
    on_off_threshold=10.0

    def __init__(self,package_type,mac,package_index,on_threshold=14.7, off_threshold=10.0,contents=None,**kwargs):
        self.package_type = package_type
        self.mac = mac.lower()
        self.uuid = self.mac.replace(":","").lower()
        self.package_index = package_index
        self.save_list=[]
        self.on_threshold = on_threshold
        self.off_threshold = off_threshold
        self.sensor_id=f"{self.uuid}_{self.package_type}_{self.package_index}"
        self.contents = contents
        print("Contents: ",self.contents)
        self.set_off_collection()

    def set_on_off_threshold(self,value):
        self.on_off_threshold = float(value)

    def set_on_threshold(self,value):
        self.on_threshold = value

    def set_off_threshold(self,value):
        self.off_threshold = value

    def set_on_collection(self):
        self.expected_state="on"
        self.set_content(self.contents['on'])
 
    def set_off_collection(self):
        self.expected_state="off"
        self.set_content(self.contents['off'])

    def set_content(self,c):
        self.content=str(c)

    def csv_filename(self,end_time):
        return f"{self.sensor_id}_{self.start_time}__{end_time}.csv"

    def get_qc_profile_filename(self):
        calibration_profile_file=f"{profilepath}{self.sensor_id}.json"
        return calibration_profile_file

    def clear_profile(self):
        f = self.get_qc_profile_filename()
        print("Clearing file: ",f)
        if os.path.isfile(f):
            os.remove(f)

    
    def save(self,ts):
        print("Writing csv..")
        fullcsvpath=f"{csvpath}{self.csv_filename(ts)}"
        print("QC CSV: ",fullcsvpath)
        print("Package Type: ",self.package_type)
        create_qc_csv(fullcsvpath,self.package_type,self.save_list)
        calibration_profile_file=self.get_qc_profile_filename()
        self.save_qc_profile(calibration_profile_file,fullcsvpath,self.content,self.start_time,ts)
        self.save_list=[]
        return calibration_profile_file

    def enable_collection(self):
        self.collect_on=True
    
    def disable_collection(self):
        self.collect_on=False

    def update_device_state(self,state):
        if (self.device_state != state):
           EventPost(QCStateChange({
               "state": state,
               "sensor_id": self.sensor_id,
               "expected_state": self.expected_state,
               "previous_state": self.device_state
           }))
           self.device_state=state
           if len(self.save_list) > 0:
               if self.device_state != self.expected_state:
                   EventPost(SensorDataCollectionError({
                       "collect_state": self.expected_state,
                        "sensor_id": self.sensor_id,
                       "state": self.device_state,
                       "msg": f"State changed to {self.device_state} after data collection started.  Expecting to stay in state {self.expected_state}"
                   }))

    def save_qc_profile(self,path,csvPath,content,stime,etime):
        profile_dict=read_json_file(path) if os.path.exists(path) else {"calibrationAttempts": [],"log": [],"header": {"host":"http://10.2.0.101:5000"}}
        entry={
                "start": stime,
                "end": etime,
                "data": {"content": content},
                "id": "642e534207c2c035033629bd",
                "sensorId": self.sensor_id,
                "createdOn": "2023-04-06T05:06:10.576Z",
                "csvPath": csvPath,
                "csvAvailable": True,
                "log": ["Record found"]
            }
        profile_dict["calibrationAttempts"].append(entry)
        write_json_file(path,profile_dict)

    def add_data(self,data):
        if not self.collect_on:
            print("---------- Collection disabled.. not saving..")
            return
        self.save_list.append(data)
        if len(self.save_list)==1:
            EventPost(SensorDataBatchCollectionStarted({
                "collected_state": self.expected_state,
                "sensor_id": self.sensor_id,
                "state": self.device_state,
                "startime":self.start_time
            }))

        if len(self.save_list)==self.collect_points:
            self.endtime = self.last_time_seen
            self.disable_collection()
            #self.save(self.last_time_seen)
            EventPost(SensorDataBatchCollectionComplete({
                "collected_state": self.expected_state,
                "state": self.device_state,
                "samples": self.collect_points,
                "endtime":self.endtime,
                "sensor_id": self.sensor_id
            }))

class VacSensorCollector(SensorCollector):

    def __init__(self,*args,contents=None,**kwargs):
        if "on_threshold" not in kwargs:
            kwargs['on_threshold'] = 14.7
        if "off_threshold" not in kwargs:
            kwargs['off_threshold'] = 10.0
        if not contents:
            contents = {
                'on': 10,
                'off': 14.7 
            }
        super().__init__(*args,contents=contents,**kwargs)
        #self.sensor_id=f"{self.mac}_{self.package_type}_{self.package_index}"
        self.collect_points=120
        self.current_bp = None
        self.set_on_off_threshold(self.contents['off'])
        self.set_on_threshold(float(self.contents['on']))
        self.set_off_threshold(float(self.contents['off']))



    def process_data_stream(self,data):

        #print("Data Value: ",token[6])
        print(f"device_state({self.device_state}) -- expected_state({self.expected_state})")
        if self.device_state == self.expected_state:
            self.start_time=int(data['timeStamp']) if not self.save_list else self.start_time
            self.last_time_seen=int(data['timeStamp'])

            data_set=','.join([ f'{float(x)}' for x in data['values']])
            self.add_data([get_datetime(int(data['timeStamp'])),int(data['index']),int(data['dataTypeIndex']),data_set,""])

    def set_state(self,data):
        value=float(data['values'][0])
        if value <= self.on_threshold + 0.5:
            self.update_device_state('on')
        elif value >= self.off_threshold - 1.0:
           self.update_device_state('off')

    def uncalibrate_vac(self,vac):
        a = 1
        b = 0

        if not self.current_bp:
            self.current_bp = 14.78
            bp_inhg = 14.78
        else:
            bp_inhg = self.current_bp * 0.0145038
        if bp_inhg > 14.7 or bp_inhg == 0.0:
            bp_inhg = 14.78

        vac_psi = bp_inhg - (vac/2.03602)
        vac_uncalibrated = (vac_psi - b)/a
        return vac_uncalibrated


    def transform_data(self,data):
        if len(data['values']) == 1:
            print("Only have one value, need to convert.")
            v = self.uncalibrate_vac(float(data['values'][0]))
            data['values'][0] = str(v) 
        elif len(data['values']) == 2:    # New format, we want the second value only.
            print("New format..")
            data['values'] = [data['values'][1]]
        else:           # Need to decide what to do here if we don't get what we expect.  Stubbed for now, but we are doing nothing. 
            pass
        return data

    def process_data(self,data):
        #print(data)
        mac_ds=data['mac'].lower()
        if data['dataType'] == 'bp':
            self.current_bp = float(data["values"][0])
        if data['sensorType'].upper()!=self.package_type or mac_ds!=self.mac:
            return
        if not self.current_bp:
            print("BP not set yet, can't adjust")
            return
        data = self.transform_data(data)
        print("Device: ",mac_ds,"Queue Length: ",len(self.save_list))
        EventPost(SensorDataPoint(data))
        self.set_state(data)
        self.process_data_stream(data)


    def format_data_stream(self,line):
        #print("Line: ",line)
        data=get_zmq_data_format2(line)
        self.process_data(data)

def create_qc_csv(path,package_type,data_list):
    with open(path, 'w') as myfile:
        writer = csv.DictWriter(myfile,fieldnames=set_header_list(),quoting=csv.QUOTE_NONNUMERIC)
        writer.writeheader()
        writer.writerows({  "datetime": row[0],
                            "type": package_type,
                            "index": int(row[1]),
                            "datatype": int(row[2]),
                            "ir":"",
                            "bp":None,
                            "hy":None,
                            "vac":row[3],
                            "pf": row[4],
                            "ps":None,
                            "value":None,
                            "x":"",
                            "y":"",
                            "z":"",
                            "var_x":None,
                            "var_y":None,
                            "var_z":None,
                            "mean_x":None,
                            "mean_y":None,
                            "mean_z":None
                        } for row in data_list)
        myfile.close()

def write_json_file(path,content):
    with open(path,'w') as f:
        json.dump(content,f)
        f.close()

def read_json_file(path):
    tmpfile={}
    with open(path) as f:
        tmpfile=json.load(f)
        f.close()
    return tmpfile

def create_qc_profile(path,csvPath,stime,etime):
    profile_dict=read_json_file(path) if os.path.exists(path) else {"calibrationAttempts": [],"log": [],"header": {"host":"http://10.2.0.101:5000"}}
    entry={
            "start": stime,
            "end": etime,
            "data": {"content": content},
            "id": "642e534207c2c035033629bd",
            "sensorId": sensor_id,
            "createdOn": "2023-04-06T05:06:10.576Z",
            "csvPath": csvPath,
            "csvAvailable": True,
            "log": ["Record found"]
        }
    profile_dict["calibrationAttempts"].append(entry)
    write_json_file(path,profile_dict)


